##
# Copyright 2009-2025 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for PETSc, implemented as an easyblock

@author: Kenneth Hoste (Ghent University)
@author: Alex Domingo (Vrije Universiteit Brussel)
"""
import glob
import os
import re

import easybuild.tools.environment as env
import easybuild.tools.toolchain as toolchain
from easybuild.easyblocks.generic.configuremake import ConfigureMake
from easybuild.framework.easyconfig import BUILD, CUSTOM
from easybuild.tools import LooseVersion
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.filetools import apply_regex_substitutions
from easybuild.tools.modules import get_software_root, get_software_version
from easybuild.tools.run import run_shell_cmd
from easybuild.tools.systemtools import get_shared_lib_ext

DEFAULT_ARCH_PATTERN = "arch*"  # default arch is 'arch-linux-c-debug'

NO_MPI_CXX_EXT_FLAGS = '-DOMPI_SKIP_MPICXX -DMPICH_SKIP_MPICXX'


class EB_PETSc(ConfigureMake):
    """Support for building and installing PETSc"""

    @staticmethod
    def extra_options():
        """Add extra config options specific to PETSc."""
        extra_vars = {
            'sourceinstall': [False, "Indicates whether a source installation should be performed", CUSTOM],
            'petsc_arch': ['', "Custom PETSC_ARCH for sourceinstall", CUSTOM],
            'shared_libs': [False, "Build shared libraries", CUSTOM],
            'with_papi': [False, "Enable PAPI support", CUSTOM],
            'papi_inc': ['/usr/include', "Path for PAPI include files", CUSTOM],
            'papi_lib': ['/usr/lib64/libpapi.so', "Path for PAPI library", CUSTOM],
            'runtest': ['test', "Make target to test build", BUILD],
            'test_parallel': [
                None,
                "Number of parallel PETSc tests launched. If unset, 'parallel' will be used",
                CUSTOM
            ],
            'download_deps_static': [[], "Dependencies that should be downloaded and installed static", CUSTOM],
            'download_deps_shared': [[], "Dependencies that should be downloaded and installed shared", CUSTOM],
            'download_deps': [[], "Dependencies that should be downloaded and installed", CUSTOM]
        }
        return ConfigureMake.extra_options(extra_vars)

    def __init__(self, *args, **kwargs):
        """Initialize PETSc specific variables."""
        super().__init__(*args, **kwargs)

        if LooseVersion(self.version) < LooseVersion("3.9"):
            raise EasyBuildError(
                f"Version {self.version} of {self.name} is unsupported. Mininum supported version is 3.9"
            )

        self.with_python = False
        self.petsc_arch = self.cfg['petsc_arch']
        self.petsc_subdir = ""
        self.petsc_root = self.installdir

        if self.cfg['sourceinstall']:
            self.build_in_installdir = True
            # sourceinstall installations place files inside an extra 'petsc-version' subdir
            # with some files located under 'petsc-version' and others under 'petsc-version/arch*'
            self.petsc_subdir = f"{self.name.lower()}-{self.version}"
            self.petsc_root = os.path.join(self.installdir, self.petsc_subdir)

        self.module_load_environment.CPATH = self.inc_dirs
        self.module_load_environment.LD_LIBRARY_PATH = self.lib_dir
        self.module_load_environment.PATH = self.bin_dir

    @property
    def arch_subdir(self):
        """Return subdir name with PETSC_ARCH or a suitable glob pattern"""
        arch_subdir = ""
        if self.cfg['sourceinstall']:
            arch_subdir = DEFAULT_ARCH_PATTERN
            if self.petsc_arch:
                arch_subdir = str(self.petsc_arch)
        return arch_subdir

    @property
    def bin_dir(self):
        """Relative path to directory with executables of PETSc"""
        # default: "lib/petsc/bin"
        # sourceinstall: "petsc-{version}/lib/petsc/bin"
        return os.path.join(self.petsc_subdir, 'lib', 'petsc', 'bin')

    @property
    def cnf_dir(self):
        """Relative path to directory with configuration executables of PETSc"""
        # default: "lib/petsc/conf"
        # sourceinstall: "petsc-{version}/{arch}/lib/petsc/conf"
        return os.path.join(self.petsc_subdir, self.arch_subdir, 'lib', 'petsc', 'conf')

    @property
    def lib_dir(self):
        """Relative path to directory with libraries of PETSc"""
        # default: "lib"
        # sourceinstall: "petsc-{version}/{arch}/lib"
        return os.path.join(self.petsc_subdir, self.arch_subdir, 'lib')

    @property
    def inc_dirs(self):
        """Relative path to directory with headers of PETSc"""
        # default: "include"
        # sourceinstall: "petsc-{version}/include" and "petsc-{version}/{arch}/include"
        if self.cfg['sourceinstall']:
            return [
                os.path.join(self.petsc_subdir, 'include'),
                os.path.join(self.petsc_subdir, self.arch_subdir, 'include'),
            ]
        return ['include']

    def prepare_step(self, *args, **kwargs):
        """Prepare build environment."""

        super().prepare_step(*args, **kwargs)

        # build with Python support if Python is loaded as a non-build (runtime) dependency
        build_deps = self.cfg.dependencies(build_only=True)
        if get_software_root('Python') and not any(x['name'] == 'Python' for x in build_deps):
            self.with_python = True
            self.module_load_environment.PYTHONPATH = self.bin_dir
            self.log.info("Python included as runtime dependency, so enabling Python support")

    def configure_step(self):
        """
        Configure PETSc by setting configure options and running configure script.

        Configure procedure is much more concise for older versions (< v3).
        """
        # make the install dir first if we are doing a download install, then keep it for the rest of the way
        deps = self.cfg["download_deps"] + self.cfg["download_deps_static"] + self.cfg["download_deps_shared"]
        if deps:
            self.log.info("Creating the installation directory before the configure.")
            self.make_installdir()
            self.cfg["keeppreviousinstall"] = True
            for dep in set(deps):
                self.cfg.update('configopts', '--download-%s=1' % dep)
            for dep in self.cfg["download_deps_static"]:
                self.cfg.update('configopts', '--download-%s-shared=0' % dep)
            for dep in self.cfg["download_deps_shared"]:
                self.cfg.update('configopts', '--download-%s-shared=1' % dep)

        # compilers
        self.cfg.update('configopts', '--with-cc="%s"' % os.getenv('CC'))
        self.cfg.update('configopts', '--with-cxx="%s" --with-c++-support' % os.getenv('CXX'))
        self.cfg.update('configopts', '--with-fc="%s"' % os.getenv('F90'))

        # compiler flags
        # Don't build with MPI c++ bindings as this leads to a hard dependency
        # on libmpi and libmpi_cxx even for C code and non-MPI code
        cxxflags = os.getenv('CXXFLAGS') + ' ' + NO_MPI_CXX_EXT_FLAGS
        self.cfg.update('configopts', '--CFLAGS="%s"' % os.getenv('CFLAGS'))
        self.cfg.update('configopts', '--CXXFLAGS="%s"' % cxxflags)
        self.cfg.update('configopts', '--FFLAGS="%s"' % os.getenv('F90FLAGS'))

        if not self.toolchain.comp_family() == toolchain.GCC:  # @UndefinedVariable
            self.cfg.update('configopts', '--with-gnu-compilers=0')

        # MPI
        if self.toolchain.options.get('usempi', None):
            self.cfg.update('configopts', '--with-mpi=1')

        # build options
        self.cfg.update('configopts', f'--with-build-step-np={self.cfg.parallel}')
        self.cfg.update('configopts', '--with-shared-libraries=%d' % self.cfg['shared_libs'])
        self.cfg.update('configopts', '--with-debugging=%d' % self.toolchain.options['debug'])
        self.cfg.update('configopts', '--with-pic=%d' % self.toolchain.options['pic'])
        self.cfg.update('configopts', '--with-x=0 --with-windows-graphics=0')

        # PAPI support
        if self.cfg['with_papi']:
            papi_inc = self.cfg['papi_inc']
            papi_inc_file = os.path.join(papi_inc, "papi.h")
            papi_lib = self.cfg['papi_lib']
            if os.path.isfile(papi_inc_file) and os.path.isfile(papi_lib):
                self.cfg.update('configopts', '--with-papi=1')
                self.cfg.update('configopts', '--with-papi-include=%s' % papi_inc)
                self.cfg.update('configopts', '--with-papi-lib=%s' % papi_lib)
            else:
                raise EasyBuildError("PAPI header (%s) and/or lib (%s) not found, can not enable PAPI support?",
                                     papi_inc_file, papi_lib)

        # Python extensions_step
        if self.with_python:

            # enable numpy support, but only if numpy is available
            res = run_shell_cmd("python -c 'import numpy'", fail_on_error=False)
            if res.exit_code == 0:
                self.cfg.update('configopts', '--with-numpy=1')

            # enable mpi4py support, but only if mpi4py is available
            res = run_shell_cmd("python -c 'import mpi4py'", fail_on_error=False)
            if res.exit_code == 0:
                with_mpi4py_opt = '--with-mpi4py'
                if self.cfg['shared_libs'] and with_mpi4py_opt not in self.cfg['configopts']:
                    self.cfg.update('configopts', '%s=1' % with_mpi4py_opt)

        # FFTW, ScaLAPACK
        deps = ["FFTW", "ScaLAPACK"]
        for dep in deps:
            libdir = os.getenv('%s_LIB_DIR' % dep.upper())
            libs = os.getenv('%s_STATIC_LIBS' % dep.upper())
            if libdir and libs:
                with_arg = "--with-%s" % dep.lower()
                self.cfg.update('configopts', '%s=1' % with_arg)
                self.cfg.update('configopts', '%s-lib=[%s/%s]' % (with_arg, libdir, libs))

                inc = os.getenv('%s_INC_DIR' % dep.upper())
                if inc:
                    self.cfg.update('configopts', '%s-include=%s' % (with_arg, inc))
            else:
                self.log.info("Missing inc/lib info, so not enabling %s support." % dep)

        # BLAS, LAPACK libraries
        bl_libdir = os.getenv('BLAS_LAPACK_LIB_DIR')
        bl_libs = os.getenv('BLAS_LAPACK_STATIC_LIBS')
        if bl_libdir and bl_libs:
            self.cfg.update('configopts', '--with-blas-lapack-lib=[%s/%s]' % (bl_libdir, bl_libs))
        else:
            raise EasyBuildError("One or more environment variables for BLAS/LAPACK not defined?")

        # additional dependencies
        # filter out deps handled seperately
        sep_deps = ['BLACS', 'BLAS', 'CMake', 'FFTW', 'LAPACK', 'numpy',
                    'mpi4py', 'papi', 'ScaLAPACK', 'SciPy-bundle', 'SuiteSparse']
        # SCOTCH has to be treated separately since they add weird postfixes
        # to library names from SCOTCH 7.0.1 or PETSc version 3.17.
        if (LooseVersion(self.version) >= LooseVersion("3.17")):
            sep_deps.append('SCOTCH')
        depfilter = [d['name'] for d in self.cfg.builddependencies()] + sep_deps

        deps = [dep['name'] for dep in self.cfg.dependencies() if not dep['name'] in depfilter]
        for dep in deps:
            if isinstance(dep, str):
                dep = (dep, dep)
            deproot = get_software_root(dep[0])
            if deproot and dep[1] == "SCOTCH":
                withdep = "--with-pt%s" % dep[1].lower()  # --with-ptscotch
                self.cfg.update('configopts', '%s=1 %s-dir=%s' % (withdep, withdep, deproot))

        # SCOTCH has to be treated separately since they add weird postfixes
        # to library names from SCOTCH 7.0.1 or PETSc version 3.17.
        scotch = get_software_root('SCOTCH')
        scotch_ver = get_software_version('SCOTCH')
        if (scotch and LooseVersion(scotch_ver) >= LooseVersion("7.0")):
            withdep = "--with-ptscotch"
            scotch_inc = [os.path.join(scotch, "include")]
            inc_spec = "-include=[%s]" % ','.join(scotch_inc)

            # For some reason there is a v3 suffix added to libptscotchparmetis
            # which is the reason for this new code;
            # note: order matters here, don't sort these alphabetically!
            req_scotch_libs = ['libptesmumps.a', 'libptscotchparmetisv3.a', 'libptscotch.a',
                               'libptscotcherr.a', 'libesmumps.a', 'libscotch.a', 'libscotcherr.a']
            scotch_libs = [os.path.join(scotch, "lib", x) for x in req_scotch_libs]
            lib_spec = "-lib=[%s]" % ','.join(scotch_libs)
            self.cfg.update('configopts', ' '.join([withdep + spec for spec in ['=1', inc_spec, lib_spec]]))

        # SuiteSparse options
        suitesparse = get_software_root('SuiteSparse')
        if suitesparse:
            withdep = "--with-suitesparse"
            # specified order of libs matters!
            ss_libs = ["UMFPACK", "KLU", "CHOLMOD", "BTF", "CCOLAMD", "COLAMD", "CAMD", "AMD"]
            # More libraries added after version 3.17
            if LooseVersion(self.version) >= LooseVersion("3.17"):
                ss_libs = ["UMFPACK", "KLU", "SPQR", "CHOLMOD", "BTF", "CCOLAMD",
                           "COLAMD", "CXSparse", "LDL", "RBio", "SLIP_LU", "CAMD", "AMD"]

            # SLIP_LU was replaced by SPEX in SuiteSparse >= 6.0
            if LooseVersion(get_software_version('SuiteSparse')) >= LooseVersion("6.0"):
                ss_libs = [x if x != "SLIP_LU" else "SPEX" for x in ss_libs]

            suitesparse_inc = os.path.join(suitesparse, "include")
            suitesparse_incs = [suitesparse_inc]
            # SuiteSparse can install its headers into a subdirectory of the include directory instead.
            suitesparse_inc_subdir = os.path.join(suitesparse_inc, 'suitesparse')
            if os.path.exists(suitesparse_inc_subdir):
                suitesparse_incs.append(suitesparse_inc_subdir)
            inc_spec = "-include=[%s]" % ','.join(suitesparse_incs)

            suitesparse_libs = [os.path.join(suitesparse, "lib", "lib%s.so" % x.replace("_", "").lower())
                                for x in ss_libs]
            lib_spec = "-lib=[%s]" % ','.join(suitesparse_libs)

            self.cfg.update('configopts', ' '.join([withdep + spec for spec in ['=1', inc_spec, lib_spec]]))

        # set PETSC_DIR for configure (env) and build_step
        petsc_dir = self.cfg['start_dir'].rstrip(os.path.sep)
        env.setvar('PETSC_DIR', petsc_dir)
        self.cfg.update('buildopts', 'PETSC_DIR=%s' % petsc_dir)

        if self.cfg['sourceinstall']:
            if self.petsc_arch:
                env.setvar('PETSC_ARCH', self.cfg['petsc_arch'])

            # run configure without --prefix (required)
            cmd = "%s ./configure %s" % (self.cfg['preconfigopts'], self.cfg['configopts'])
            res = run_shell_cmd(cmd)
            out = res.output
        else:
            out = super().configure_step()

        # check for errors in configure
        error_regexp = re.compile("ERROR")
        if error_regexp.search(out):
            raise EasyBuildError("Error(s) detected in configure output!")

        if self.cfg['sourceinstall']:
            # figure out PETSC_ARCH setting
            petsc_arch_regex = re.compile(r"^\s*PETSC_ARCH:\s*(\S+)$", re.M)
            res = petsc_arch_regex.search(out)
            if res:
                self.petsc_arch = res.group(1)
                self.cfg.update('buildopts', 'PETSC_ARCH=%s' % self.petsc_arch)
            else:
                raise EasyBuildError("Failed to determine PETSC_ARCH setting.")

        # Make sure to set test_parallel before self.cfg.parallel is set to 1
        if self.cfg['test_parallel'] is None:
            self.cfg['test_parallel'] = self.cfg.parallel

        # PETSc make does not accept -j
        # to control parallel build, we need to specify MAKE_NP=... as argument to 'make' command
        self.cfg.update('buildopts', f"MAKE_NP={self.cfg.parallel}")
        self.cfg.parallel = 1

    # default make should be fine

    def test_step(self):
        """
        Test the compilation
        """

        # Each PETSc test may use multiple threads, so running "self.cfg.parallel" of them may lead to
        # some oversubscription every now and again. Not a big deal, but if needed a reduced parallelism
        # can be specified with test_parallel - and it takes priority
        paracmd = ''
        self.log.info("In test_step: %s" % self.cfg['test_parallel'])
        if self.cfg['test_parallel'] > 1:
            paracmd = f"-j {self.cfg['test_parallel']}"

        if self.cfg['runtest']:
            cmd = "%s make %s %s %s" % (self.cfg['pretestopts'], paracmd, self.cfg['runtest'], self.cfg['testopts'])
            res = run_shell_cmd(cmd)
            out = res.output

            return out

    def install_step(self):
        """
        Install using make install (for non-source installations)
        """
        if not self.cfg['sourceinstall']:
            super().install_step()

        # Remove MPI-CXX flags added during configure to prevent them from being passed to consumers of PETsc
        petsc_variables_path = os.path.join(self.petsc_root, 'lib', 'petsc', 'conf', 'petscvariables')
        if os.path.isfile(petsc_variables_path):
            fix = (r'^(CXX_FLAGS|CXX_LINKER_FLAGS|CONFIGURE_OPTIONS)( = .*)%s(.*)$' % NO_MPI_CXX_EXT_FLAGS,
                   r'\1\2\3')
            apply_regex_substitutions(petsc_variables_path, [fix])

    def make_module_extra(self):
        """Set PETSc specific environment variables (PETSC_DIR, PETSC_ARCH)."""
        txt = super().make_module_extra()

        txt += self.module_generator.set_environment('PETSC_DIR', self.petsc_root)
        if self.cfg['sourceinstall']:
            txt += self.module_generator.set_environment('PETSC_ARCH', self.petsc_arch)

        return txt

    def sanity_check_step(self):
        """Custom sanity check for PETSc"""

        libext = 'a'
        if self.cfg['shared_libs']:
            libext = get_shared_lib_ext()

        if self.cfg['sourceinstall'] and not self.petsc_arch:
            # in module only runs we need to determine PETSC_ARCH from the installation files
            for arch_subdir in glob.glob(os.path.join(self.petsc_root, '*', 'lib', 'petsc')):
                self.petsc_arch = arch_subdir.split(os.sep)[-3]

        custom_paths = {
            'files': [
                os.path.join(self.bin_dir, 'petscmpiexec'),
                os.path.join(self.cnf_dir, 'petscvariables'),
                os.path.join(self.lib_dir, f'libpetsc.{libext}'),
            ],
            'dirs': self.inc_dirs,
        }

        custom_commands = []
        if self.with_python:
            custom_commands.append("python -m PetscBinaryIO --help")

        super().sanity_check_step(custom_paths=custom_paths, custom_commands=custom_commands)



# #
# Copyright 2009-2025 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
# #
"""
EasyBuild support for installing the Intel MPI library, implemented as an easyblock

@author: Stijn De Weirdt (Ghent University)
@author: Dries Verdegem (Ghent University)
@author: Kenneth Hoste (Ghent University)
@author: Pieter De Baets (Ghent University)
@author: Jens Timmerman (Ghent University)
@author: Damian Alvarez (Forschungszentrum Juelich GmbH)
@author: Alex Domingo (Vrije Universiteit Brussel)
@author: Jan Andre Reuter (Forschungszentrum Juelich GmbH)
"""
import glob
import os
import tempfile
from easybuild.tools import LooseVersion

import easybuild.tools.toolchain as toolchain
from easybuild.easyblocks.generic.intelbase import IntelBase
from easybuild.framework.easyconfig import CUSTOM
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.config import build_option
from easybuild.tools.filetools import apply_regex_substitutions, change_dir, copy_files, extract_file
from easybuild.tools.filetools import mkdir, remove, get_cwd
from easybuild.tools.modules import MODULE_LOAD_ENV_HEADERS, get_software_root, get_software_version
from easybuild.tools.run import run_shell_cmd
from easybuild.tools.systemtools import get_shared_lib_ext
from easybuild.tools.toolchain.mpi import get_mpi_cmd_template


class EB_impi(IntelBase):
    """
    Support for installing Intel MPI library
    - minimum version suported: 2018.x
    """
    @staticmethod
    def extra_options():
        extra_vars = {
            'libfabric_configopts': ['', 'Configure options for the provided libfabric', CUSTOM],
            'libfabric_rebuild': [True, 'Try to rebuild internal libfabric instead of using provided binary', CUSTOM],
            'ofi_internal': [True, 'Use internal shipped libfabric instead of external libfabric', CUSTOM],
            'set_mpi_wrappers_compiler': [False, 'Override default compiler used by MPI wrapper commands', CUSTOM],
            'set_mpi_wrapper_aliases_gcc': [False, 'Set compiler for mpigcc/mpigxx via aliases', CUSTOM],
            'set_mpi_wrapper_aliases_intel': [False, 'Set compiler for mpiicc/mpiicpc/mpiifort via aliases', CUSTOM],
            'set_mpi_wrappers_all': [False, 'Set (default) compiler for all MPI wrapper commands', CUSTOM],
            'rebuild_f08_bindings': [False, 'Rebuild and replace the Fortran 2008 bindings.'
                                            'Not built by default due to backwards compatibility.', CUSTOM],
        }
        return IntelBase.extra_options(extra_vars)

    def prepare_step(self, *args, **kwargs):
        kwargs['requires_runtime_license'] = False
        super().prepare_step(*args, **kwargs)

    def install_step(self):
        """
        Actual installation
        - create silent cfg file
        - execute command
        """
        impiver = LooseVersion(self.version)

        if impiver < LooseVersion('2018'):
            raise EasyBuildError(
                f"Version {self.version} of {self.name} is unsupported. Mininum supported version is 2018.0."
            )

        if impiver >= LooseVersion('2021'):
            super().install_step()
        else:
            # impi starting from version 4.0.1.x uses standard installation procedure.
            silent_cfg_names_map = {}
            super().install_step(silent_cfg_names_map=silent_cfg_names_map)
            # since v5.0.1 installers create impi/<version> subdir, so stuff needs to be moved afterwards
            super().move_after_install()

        # recompile libfabric (if requested)
        # some Intel MPI versions (like 2019 update 6) no longer ship libfabric sources
        libfabric_path = os.path.join(self.installdir, 'libfabric')
        if impiver >= LooseVersion('2019') and self.cfg['libfabric_rebuild']:
            if self.cfg['ofi_internal']:
                libfabric_src_tgz_fn = 'src.tgz'
                if os.path.exists(os.path.join(libfabric_path, libfabric_src_tgz_fn)):
                    change_dir(libfabric_path)
                    srcdir = extract_file(libfabric_src_tgz_fn, os.getcwd(), change_into_dir=False)
                    change_dir(srcdir)
                    libfabric_installpath = os.path.join(self.installdir, 'intel64', 'libfabric')

                    make = 'make'
                    if self.cfg.parallel > 1:
                        make += f' -j {self.cfg.parallel}'

                    cmds = [
                        f"./configure --prefix={libfabric_installpath} {self.cfg['libfabric_configopts']}",
                        make,
                        "make install",
                    ]
                    for cmd in cmds:
                        run_shell_cmd(cmd)
                else:
                    self.log.info("Rebuild of libfabric is requested, but %s does not exist, so skipping...",
                                  libfabric_src_tgz_fn)
            else:
                raise EasyBuildError("Rebuild of libfabric is requested, but ofi_internal is set to False.")

    def rebuild_f08_bindings(self):
        """
        Rebuild Fortran 2008 bindings
        """
        # For more information, see:
        # https://community.intel.com/t5/Intel-MPI-Library/MPI-f08-with-polymorphic-argument-CLASS/m-p/1590421
        # Check if the binding tarball exists within the installation
        bindings_path = os.path.join(self.installdir, 'mpi', 'latest', 'opt', 'mpi', 'binding')
        bindings_tarball = os.path.join(bindings_path, 'intel-mpi-binding-kit.tar.gz')
        if not os.path.exists(bindings_tarball):
            raise EasyBuildError(
                f"Requested to rebuild Fortran 2008 bindings, but the bindings tarball in {bindings_tarball} "
                f"does not exist.")
        self.log.info(f"Found bindings tarball at {bindings_tarball}. Rebuilding Fortran 2008 bindings.")
        # Determine compilers to build with
        if self.toolchain.is_system_toolchain():
            f90_compiler = 'gfortran'
        else:
            f90_compiler = self.toolchain.COMPILER_F90
        # Extract the tarball
        build_dir = tempfile.mkdtemp(prefix='rebuild-bindings-', dir=self.builddir)
        run_shell_cmd(f"tar -xzf {bindings_tarball} -C {build_dir}")
        # Build the bindings
        change_dir(os.path.join(build_dir, 'f08'))
        mpi_latest_dir = os.path.join(self.installdir, 'mpi', 'latest')
        run_shell_cmd(f"make MPI_INST={mpi_latest_dir} F90={f90_compiler} NAME={f90_compiler}")
        change_dir(os.path.join(get_cwd(), 'include', f'{f90_compiler}'))
        include_mpi_dir = os.path.join(mpi_latest_dir, 'include', 'mpi')
        initial_module_files = glob.glob(os.path.join(include_mpi_dir, '*.mod'))
        written_module_files = glob.glob("*.mod")
        # Preserve the original module files for people to use
        include_mpi_originals_dir = os.path.join(include_mpi_dir, 'originals')
        mkdir(include_mpi_originals_dir, parents=True)
        copy_files(initial_module_files, include_mpi_originals_dir)
        # Copy the new module files
        copy_files(written_module_files, include_mpi_dir)
        # Cleanup
        remove(build_dir)
        change_dir(self.installdir)

    def post_processing_step(self):
        """Custom post install step for IMPI, fix broken env scripts after moving installed files."""
        super().post_processing_step()

        impiver = LooseVersion(self.version)

        if impiver >= LooseVersion('2021'):
            self.log.info("No post-install action for impi v%s", self.version)
        else:
            script_paths = [os.path.join('intel64', 'bin')]
            # fix broken env scripts after the move
            regex_subs = [(r"^setenv I_MPI_ROOT.*", r"setenv I_MPI_ROOT %s" % self.installdir)]
            for script in [os.path.join(script_path, 'mpivars.csh') for script_path in script_paths]:
                apply_regex_substitutions(os.path.join(self.installdir, script), regex_subs)
            regex_subs = [(r"^(\s*)I_MPI_ROOT=[^;\n]*", r"\1I_MPI_ROOT=%s" % self.installdir)]
            for script in [os.path.join(script_path, 'mpivars.sh') for script_path in script_paths]:
                apply_regex_substitutions(os.path.join(self.installdir, script), regex_subs)

            # fix 'prefix=' in compiler wrapper scripts after moving installation (see install_step)
            wrappers = ['mpif77', 'mpif90', 'mpigcc', 'mpigxx', 'mpiicc', 'mpiicpc', 'mpiifort']
            regex_subs = [(r"^prefix=.*", r"prefix=%s" % self.installdir)]
            for script_dir in script_paths:
                for wrapper in wrappers:
                    wrapper_path = os.path.join(self.installdir, script_dir, wrapper)
                    if os.path.exists(wrapper_path):
                        apply_regex_substitutions(wrapper_path, regex_subs)

        if self.cfg['rebuild_f08_bindings']:
            self.rebuild_f08_bindings()

    def sanity_check_step(self):
        """Custom sanity check paths for IMPI."""

        impi_ver = LooseVersion(self.version)

        suff = '64'

        mpi_mods = ['mpi.mod', 'mpi_base.mod', 'mpi_constants.mod', 'mpi_sizeofs.mod']

        if impi_ver >= LooseVersion('2021'):
            mpi_subdir = self.get_versioned_subdir('mpi')
            bin_dir = os.path.join(mpi_subdir, 'bin')
            include_dir = os.path.join(mpi_subdir, 'include')
            lib_dir = os.path.join(mpi_subdir, 'lib')
            if impi_ver < LooseVersion('2021.11'):
                lib_dir = os.path.join(lib_dir, 'release')
        elif impi_ver >= LooseVersion('2019'):
            bin_dir = os.path.join('intel64', 'bin')
            include_dir = os.path.join('intel64', 'include')
            lib_dir = os.path.join('intel64', 'lib', 'release')
        else:
            bin_dir = 'bin%s' % suff
            include_dir = 'include%s' % suff
            lib_dir = 'lib%s' % suff
            mpi_mods.extend(['i_malloc.h'])

        mpi_mods_dir = include_dir
        if impi_ver >= LooseVersion('2021.11'):
            mpi_mods_dir = os.path.join(mpi_mods_dir, 'mpi')

        shlib_ext = get_shared_lib_ext()
        custom_paths = {
            'files': [os.path.join(bin_dir, 'mpi%s' % x) for x in ['icc', 'icpc', 'ifort']] +
            [os.path.join(include_dir, 'mpi%s.h' % x) for x in ['cxx', 'f', '', 'o', 'of']] +
            [os.path.join(mpi_mods_dir, x) for x in mpi_mods] +
            [os.path.join(lib_dir, 'libmpi.%s' % shlib_ext)] +
            [os.path.join(lib_dir, 'libmpi.a')],
            'dirs': [],
        }

        custom_commands = []

        if build_option('mpi_tests'):
            # Add minimal test program to sanity checks
            if build_option('sanity_check_only'):
                # When only running the sanity check we need to manually make sure that
                # variables for compilers and parallelism have been set
                self.set_parallel()
                self.prepare_step(start_dir=False)

                impi_testexe = os.path.join(tempfile.mkdtemp(), 'mpi_test')
            else:
                impi_testexe = os.path.join(self.builddir, 'mpi_test')

            if impi_ver >= LooseVersion('2021'):
                impi_testsrc = os.path.join(self.installdir, self.get_versioned_subdir('mpi'))
                if impi_ver >= LooseVersion('2021.11'):
                    impi_testsrc = os.path.join(impi_testsrc, 'opt', 'mpi')
                impi_testsrc = os.path.join(impi_testsrc, 'test', 'test.c')
            else:
                impi_testsrc = os.path.join(self.installdir, 'test', 'test.c')

            self.log.info("Adding minimal MPI test program to sanity checks: %s", impi_testsrc)

            # Build test program with appropriate compiler from current toolchain
            build_cmd = "mpicc -cc=%s %s -o %s" % (os.getenv('CC'), impi_testsrc, impi_testexe)

            # Execute test program with appropriate MPI executable for target toolchain
            params = {'nr_ranks': self.cfg.parallel, 'cmd': impi_testexe}
            mpi_cmd_tmpl, params = get_mpi_cmd_template(toolchain.INTELMPI, params, mpi_version=self.version)

            custom_commands.extend([
                build_cmd,  # build test program
                mpi_cmd_tmpl % params,  # run test program
            ])

        super().sanity_check_step(custom_paths=custom_paths, custom_commands=custom_commands)

    def make_module_step(self, *args, **kwargs):
        """
        Set paths for module load environment based on the actual installation files
        """
        manpath = 'man'
        fi_provider_path = None
        mic_library_path = None

        impi_ver = LooseVersion(self.version)
        if impi_ver >= LooseVersion('2021'):
            mpi_subdir = self.get_versioned_subdir('mpi')
            path_dirs = [
                os.path.join(mpi_subdir, 'bin'),
                os.path.join(mpi_subdir, 'libfabric', 'bin'),
            ]
            lib_dirs = [
                os.path.join(mpi_subdir, 'lib'),
                os.path.join(mpi_subdir, 'libfabric', 'lib'),
            ]
            if impi_ver < LooseVersion('2021.11'):
                lib_dirs.insert(1, os.path.join(mpi_subdir, 'lib', 'release'))
            include_dirs = [os.path.join(mpi_subdir, 'include')]

            if impi_ver >= LooseVersion('2021.11'):
                manpath = os.path.join(mpi_subdir, 'share', 'man')
            else:
                manpath = os.path.join(mpi_subdir, 'man')

            if self.cfg['ofi_internal']:
                lib_dirs.append(os.path.join(mpi_subdir, 'libfabric', 'lib'))
                path_dirs.append(os.path.join(mpi_subdir, 'libfabric', 'bin'))
                fi_provider_path = [os.path.join(mpi_subdir, 'libfabric', 'lib', 'prov')]

        elif impi_ver >= LooseVersion('2019'):
            path_dirs = [os.path.join('intel64', 'bin')]
            # The "release" library is default in v2019. Give it precedence over intel64/lib.
            # (remember paths are *prepended*, so the last path in the list has highest priority)
            lib_dirs = [
                os.path.join('intel64', 'lib'),
                os.path.join('intel64', 'lib', 'release'),
            ]
            include_dirs = [os.path.join('intel64', 'include')]

            if self.cfg['ofi_internal']:
                lib_dirs.append(os.path.join('intel64', 'libfabric', 'lib'))
                path_dirs.append(os.path.join('intel64', 'libfabric', 'bin'))
                fi_provider_path = [os.path.join('intel64', 'libfabric', 'lib', 'prov')]

        else:
            path_dirs = [os.path.join('bin', 'intel64'), 'bin64']
            lib_dirs = [os.path.join('lib', 'em64t'), 'lib64']
            include_dirs = ['include64']
            mic_library_path = [os.path.join('mic', 'lib')]

        self.module_load_environment.PATH = path_dirs
        self.module_load_environment.LD_LIBRARY_PATH = lib_dirs
        self.module_load_environment.LIBRARY_PATH = lib_dirs
        self.module_load_environment.MANPATH = [manpath]
        if fi_provider_path is not None:
            self.module_load_environment.FI_PROVIDER_PATH = fi_provider_path
        if mic_library_path is not None:
            self.module_load_environment.MIC_LD_LIBRARY_PATH = mic_library_path

        # include paths to headers (e.g. CPATH)
        self.module_load_environment.set_alias_vars(MODULE_LOAD_ENV_HEADERS, include_dirs)

        return super().make_module_step(*args, **kwargs)

    def make_module_extra(self, *args, **kwargs):
        """Overwritten from Application to add extra txt"""

        if LooseVersion(self.version) >= LooseVersion('2021'):
            mpiroot = os.path.join(self.installdir, self.get_versioned_subdir('mpi'))
        else:
            mpiroot = self.installdir

        txt = super().make_module_extra(*args, **kwargs)
        txt += self.module_generator.set_environment('I_MPI_ROOT', mpiroot)
        if self.cfg['set_mpi_wrappers_compiler'] or self.cfg['set_mpi_wrappers_all']:
            for var in ['CC', 'CXX', 'F77', 'F90', 'FC']:
                if var == 'FC':
                    # $FC isn't defined by EasyBuild framework, so use $F90 instead
                    src_var = 'F90'
                else:
                    src_var = var

                target_var = 'I_MPI_%s' % var

                val = os.getenv(src_var)
                if val:
                    txt += self.module_generator.set_environment(target_var, val)
                else:
                    raise EasyBuildError("Environment variable $%s not set, can't define $%s", src_var, target_var)

        if self.cfg['set_mpi_wrapper_aliases_gcc'] or self.cfg['set_mpi_wrappers_all']:
            # force mpigcc/mpigxx to use GCC compilers, as would be expected based on their name
            txt += self.module_generator.set_alias('mpigcc', 'mpigcc -cc=gcc')
            txt += self.module_generator.set_alias('mpigxx', 'mpigxx -cxx=g++')

        if self.cfg['set_mpi_wrapper_aliases_intel'] or self.cfg['set_mpi_wrappers_all']:
            # do the same for mpiicc/mpiipc/mpiifort to be consistent, even if they may not exist
            if (get_software_root('intel-compilers') and
                    LooseVersion(get_software_version('intel-compilers')) >= LooseVersion('2024')):
                txt += self.module_generator.set_alias('mpiicc', 'mpiicc -cc=icx')
                txt += self.module_generator.set_alias('mpiicpc', 'mpiicpc -cxx=icpx')
            else:
                txt += self.module_generator.set_alias('mpiicc', 'mpiicc -cc=icc')
                txt += self.module_generator.set_alias('mpiicpc', 'mpiicpc -cxx=icpc')
            # -fc also works, but -f90 takes precedence
            txt += self.module_generator.set_alias('mpiifort', 'mpiifort -f90=ifort')

            if LooseVersion(self.version) >= LooseVersion('2021.11'):
                txt += self.module_generator.set_alias('mpiicx', 'mpiicx -cc=icx')
                txt += self.module_generator.set_alias('mpiicpx', 'mpiicpx -cxx=icpx')
                txt += self.module_generator.set_alias('mpiifx', 'mpiifx -f90=ifx')

        # set environment variable UCX_TLS to 'all', this works in all hardware configurations
        # needed with UCX regardless of the transports available (even without a Mellanox HCA)
        # more information in easybuilders/easybuild-easyblocks#2253
        if get_software_root('UCX'):
            # do not overwrite settings in the easyconfig
            if 'UCX_TLS' not in self.cfg['modextravars']:
                txt += self.module_generator.set_environment('UCX_TLS', 'all')

        return txt



##
# Copyright 2009-2025 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for installing a bundle of modules, implemented as a generic easyblock

@author: Stijn De Weirdt (Ghent University)
@author: Dries Verdegem (Ghent University)
@author: Kenneth Hoste (Ghent University)
@author: Pieter De Baets (Ghent University)
@author: Jens Timmerman (Ghent University)
@author: Jasper Grimm (University of York)
@author: Jan Andre Reuter (Juelich Supercomputing Centre)
"""
import copy
import os

import easybuild.tools.environment as env
from easybuild.framework.easyblock import EasyBlock
from easybuild.framework.easyconfig import CUSTOM
from easybuild.framework.easyconfig.easyconfig import get_easyblock_class
from easybuild.tools.build_log import EasyBuildError, print_msg
from easybuild.tools.modules import get_software_root, get_software_version
from easybuild.tools.utilities import nub


class Bundle(EasyBlock):
    """
    Bundle of modules: only generate module files, nothing to build/install
    """

    @staticmethod
    def extra_options(extra_vars=None):
        """Easyconfig parameters specific to bundles."""
        if extra_vars is None:
            extra_vars = {}
        extra_vars.update({
            'altroot': [None, "Software name of dependency to use to define $EBROOT for this bundle", CUSTOM],
            'altversion': [None, "Software name of dependency to use to define $EBVERSION for this bundle", CUSTOM],
            'default_component_specs': [{}, "Default specs to use for every component", CUSTOM],
            'components': [(), "List of components to install: tuples w/ name, version and easyblock to use", CUSTOM],
            'sanity_check_components': [[], "List of components for which to run sanity checks", CUSTOM],
            'sanity_check_all_components': [False, "Enable sanity checks for all components", CUSTOM],
            'default_easyblock': [None, "Default easyblock to use for components", CUSTOM],
        })
        return EasyBlock.extra_options(extra_vars)

    def __init__(self, *args, **kwargs):
        """Initialize easyblock."""
        super().__init__(*args, **kwargs)
        self.altroot = None
        self.altversion = None

        # list of EasyConfig instances and their EasyBlocks for components
        self.comp_instances = []

        # list of EasyConfig instances of components for which to run sanity checks
        self.comp_cfgs_sanity_check = []

        check_for_sources = getattr(self, 'check_for_sources', True)
        # list of sources for bundle itself *must* be empty (unless overridden by subclass)
        if check_for_sources:
            if self.cfg.get_ref('sources'):
                raise EasyBuildError("List of sources for bundle itself must be empty, found %s", self.cfg['sources'])
            if self.cfg.get_ref('patches'):
                raise EasyBuildError("List of patches for bundle itself must be empty, found %s", self.cfg['patches'])

        # copy EasyConfig instance before we make changes to it
        # (like adding component sources to top-level sources easyconfig parameter)
        self.cfg = self.cfg.copy()

        # disable templating to avoid premature resolving of template values
        # Note that self.cfg.update also resolves templates!
        with self.cfg.disable_templating():
            # list of checksums for patches (must be included after checksums for sources)
            checksums_patches = []

            if self.cfg['sanity_check_components'] and self.cfg['sanity_check_all_components']:
                raise EasyBuildError("sanity_check_components and sanity_check_all_components"
                                     "cannot be enabled together")

            # backup and reset general sanity checks from main body of ec,
            # if component-specific sanity checks are enabled necessary to avoid:
            # - duplicating the general sanity check across all components running sanity checks
            # - general sanity checks taking precedence over those defined in a component's easyblock
            self.backup_sanity_paths = self.cfg['sanity_check_paths']
            self.backup_sanity_cmds = self.cfg['sanity_check_commands']
            if self.cfg['sanity_check_components'] or self.cfg['sanity_check_all_components']:
                # reset general sanity checks, to be restored later
                self.cfg['sanity_check_paths'] = {}
                self.cfg['sanity_check_commands'] = {}
            components = self.cfg['components']

            for comp in components:
                comp_name, comp_version, comp_specs = comp[0], comp[1], {}
                if len(comp) == 3:
                    comp_specs = comp[2]

                comp_cfg = self.cfg.copy()

                comp_cfg['name'] = comp_name
                comp_cfg['version'] = comp_version

                # determine easyblock to use for this component
                # - if an easyblock is specified explicitely, that will be used
                # - if not, a software-specific easyblock will be considered by get_easyblock_class
                # - if no easyblock was found, default_easyblock is considered
                comp_easyblock = comp_specs.get('easyblock')
                easyblock_class = get_easyblock_class(comp_easyblock, name=comp_name, error_on_missing_easyblock=False)
                if easyblock_class is None:
                    if self.cfg['default_easyblock']:
                        easyblock = self.cfg['default_easyblock']
                        easyblock_class = get_easyblock_class(easyblock)

                    if easyblock_class is None:
                        raise EasyBuildError("No easyblock found for component %s v%s", comp_name, comp_version)
                    self.log.info("Using default easyblock %s for component %s", easyblock, comp_name)
                else:
                    easyblock = easyblock_class.__name__
                    self.log.info("Using easyblock %s for component %s", easyblock, comp_name)

                if easyblock == 'Bundle':
                    raise EasyBuildError("The Bundle easyblock can not be used to install components in a bundle")

                comp_cfg.easyblock = easyblock_class

                # make sure that extra easyconfig parameters are known, so they can be set
                extra_opts = comp_cfg.easyblock.extra_options()
                comp_cfg.extend_params(copy.deepcopy(extra_opts))

                comp_cfg.generate_template_values()

                # do not inherit easyblock to use from parent
                # (since that would result in an infinite loop in install_step)
                comp_cfg['easyblock'] = None

                # reset list of sources/source_urls/checksums
                comp_cfg['sources'] = comp_cfg['source_urls'] = comp_cfg['checksums'] = comp_cfg['patches'] = []

                for key in self.cfg['default_component_specs']:
                    comp_cfg[key] = self.cfg['default_component_specs'][key]

                for key in comp_specs:
                    comp_cfg[key] = comp_specs[key]

                # Don't require that all template values can be resolved at this point but still resolve them.
                # This is important to ensure that template values like %(name)s and %(version)s
                # are correctly resolved with the component name/version before values are copied over to self.cfg
                with comp_cfg.allow_unresolved_templates():
                    comp_sources = comp_cfg['sources']
                    comp_source_urls = comp_cfg['source_urls']
                if not comp_sources:
                    raise EasyBuildError("No sources specification for component %s v%s", comp_name, comp_version)
                # If per-component source URLs are provided, attach them directly to the relevant sources
                if comp_source_urls:
                    for source in comp_sources:
                        if isinstance(source, str):
                            self.cfg.update('sources', [{'filename': source, 'source_urls': comp_source_urls[:]}])
                        elif isinstance(source, dict):
                            # Update source_urls in the 'source' dict to use the one for the components
                            # (if it doesn't already exist)
                            if 'source_urls' not in source:
                                source['source_urls'] = comp_source_urls[:]
                            self.cfg.update('sources', [source])
                        else:
                            raise EasyBuildError("Source %s for component %s is neither a string nor a dict, cannot "
                                                 "process it.", source, comp_cfg['name'])
                else:
                    # add component sources to list of sources
                    self.cfg.update('sources', comp_sources)

                comp_checksums = comp_cfg['checksums']
                if comp_checksums:
                    src_cnt = len(comp_sources)

                    # add per-component checksums for sources to list of checksums
                    self.cfg.update('checksums', comp_checksums[:src_cnt])

                    # add per-component checksums for patches to list of checksums for patches
                    checksums_patches.extend(comp_checksums[src_cnt:])

                with comp_cfg.allow_unresolved_templates():
                    comp_patches = comp_cfg['patches']
                if comp_patches:
                    self.cfg.update('patches', comp_patches)

                self.comp_instances.append((comp_cfg, comp_cfg.easyblock(comp_cfg, logfile=self.logfile)))

            self.cfg.update('checksums', checksums_patches)

        # restore general sanity checks if using component-specific sanity checks
        if self.cfg['sanity_check_components'] or self.cfg['sanity_check_all_components']:
            self.cfg['sanity_check_paths'] = self.backup_sanity_paths
            self.cfg['sanity_check_commands'] = self.backup_sanity_cmds

    def check_checksums(self):
        """
        Check whether a SHA256 checksum is available for all sources & patches (incl. extensions).

        :return: list of strings describing checksum issues (missing checksums, wrong checksum type, etc.)
        """
        checksum_issues = super().check_checksums()

        for comp, _ in self.comp_instances:
            checksum_issues.extend(self.check_checksums_for(comp, sub="of component %s" % comp['name']))

        return checksum_issues

    def patch_step(self):
        """Patch step must be a no-op for bundle, since there are no top-level sources/patches."""
        pass

    def get_altroot_and_altversion(self):
        """Get altroot and altversion, if they are defined"""
        altroot = None
        if self.cfg['altroot']:
            altroot = get_software_root(self.cfg['altroot'])
        altversion = None
        if self.cfg['altversion']:
            altversion = get_software_version(self.cfg['altversion'])
        return altroot, altversion

    def configure_step(self):
        """Collect altroot/altversion info."""
        self.altroot, self.altversion = self.get_altroot_and_altversion()

    def build_step(self):
        """Do nothing."""
        pass

    def install_step(self):
        """Install components, if specified."""
        comp_cnt = len(self.cfg['components'])
        for idx, (cfg, comp) in enumerate(self.comp_instances):

            print_msg("installing bundle component %s v%s (%d/%d)..." %
                      (cfg['name'], cfg['version'], idx + 1, comp_cnt))
            self.log.info("Installing component %s v%s using easyblock %s", cfg['name'], cfg['version'], cfg.easyblock)

            # correct build/install dirs
            comp.builddir = self.builddir
            comp.install_subdir, comp.installdir = self.install_subdir, self.installdir

            # make sure we can build in parallel
            comp.set_parallel()

            # figure out correct start directory
            comp.guess_start_dir()

            # need to run fetch_patches to ensure per-component patches are applied
            comp.fetch_patches()

            comp.src = []

            # find matching entries in self.src for this component
            with comp.cfg.allow_unresolved_templates():
                comp_sources = comp.cfg['sources']
            for source in comp_sources:
                if isinstance(source, str):
                    comp_src_fn = source
                elif isinstance(source, dict):
                    if 'filename' in source:
                        comp_src_fn = source['filename']
                    else:
                        raise EasyBuildError("Encountered source file specified as dict without 'filename': %s", source)
                else:
                    raise EasyBuildError("Specification of unknown type for source file: %s", source)

                found = False
                for src in self.src:
                    if src['name'] == comp_src_fn:
                        self.log.info("Found spec for source %s for component %s: %s", comp_src_fn, comp.name, src)
                        comp.src.append(src)
                        found = True
                        break
                if not found:
                    raise EasyBuildError("Failed to find spec for source %s for component %s", comp_src_fn, comp.name)

                # location of first unpacked source is used to determine where to apply patch(es)
                comp.src[-1]['finalpath'] = comp.cfg['start_dir']

            # check if sanity checks are enabled for the component
            if self.cfg['sanity_check_all_components'] or comp.cfg['name'] in self.cfg['sanity_check_components']:
                self.comp_cfgs_sanity_check.append(comp)

            # run relevant steps
            for step_name in ['patch', 'configure', 'build', 'install']:
                if step_name in cfg['skipsteps']:
                    comp.log.info("Skipping '%s' step for component %s v%s", step_name, cfg['name'], cfg['version'])
                else:
                    comp.run_step(step_name, [lambda x: getattr(x, '%s_step' % step_name)])

            if comp.make_module_req_guess.__qualname__ != 'EasyBlock.make_module_req_guess':
                depr_msg = f"Easyblock used to install component {comp.name} still uses make_module_req_guess"
                self.log.deprecated(depr_msg, '6.0')
                # update environment to ensure stuff provided by former components can be picked up by latter components
                # once the installation is finalised, this is handled by the generated module
                reqs = comp.make_module_req_guess()
                for envvar in reqs:
                    curr_val = os.getenv(envvar, '')
                    curr_paths = curr_val.split(os.pathsep)
                    for subdir in reqs[envvar]:
                        path = os.path.join(self.installdir, subdir)
                        if path not in curr_paths:
                            if curr_val:
                                new_val = '%s:%s' % (path, curr_val)
                            else:
                                new_val = path
                            env.setvar(envvar, new_val)
            else:
                # Update current environment with component environment to ensure stuff provided
                # by this component can be picked up by installation of subsequent components
                # - this is a stripped down version of EasyBlock.make_module_req for fake modules
                # - once bundle installation is complete, this is handled by the generated module as usual
                for mod_envar, mod_paths in comp.module_load_environment.items():
                    # expand glob patterns in module load environment to existing absolute paths
                    mod_expand = mod_paths.expand_paths(self.installdir)
                    mod_expand = [os.path.join(self.installdir, path) for path in mod_expand]
                    # prepend to current environment variable if new stuff added to installation
                    curr_env = os.getenv(mod_envar, '')
                    curr_paths = [path for path in curr_env.split(os.pathsep) if path]
                    new_paths = nub(mod_expand + curr_paths)
                    new_env = os.pathsep.join(new_paths)
                    if new_env and new_env != curr_env:
                        env.setvar(mod_envar, new_env)

    def make_module_step(self, *args, **kwargs):
        """
        Set module requirements from all components, e.g. $PATH, etc.
        During the install step, we only set these requirements temporarily.
        Later on when building the module, those paths are not considered.
        Therefore, iterate through all the components again and gather
        the requirements.

        Do not remove duplicates or check for existence of folders,
        as this is done in the generic EasyBlock while creating
        the module file already.
        """
        for cfg, comp in self.comp_instances:
            self.log.info("Gathering module paths for component %s v%s", cfg['name'], cfg['version'])

            # take into account that easyblock used for component may not be migrated yet to module_load_environment
            if comp.make_module_req_guess.__qualname__ != 'EasyBlock.make_module_req_guess':

                depr_msg = f"Easyblock used to install component {cfg['name']} still uses make_module_req_guess"
                self.log.deprecated(depr_msg, '6.0')

                reqs = comp.make_module_req_guess()

                # Try-except block to fail with an easily understandable error message.
                # This should only trigger when an EasyBlock returns non-dict module requirements
                # for make_module_req_guess() which should then be fixed in the components EasyBlock.
                try:
                    for key, value in sorted(reqs.items()):
                        if key in self.module_load_environment:
                            getattr(self.module_load_environment, key).extend(value)
                        else:
                            setattr(self.module_load_environment, key, value)
                except AttributeError:
                    raise EasyBuildError("Cannot process module requirements of bundle component %s v%s",
                                         cfg['name'], cfg['version'])
            else:
                for env_var_name, env_var_val in comp.module_load_environment.items():
                    if env_var_name in self.module_load_environment:
                        getattr(self.module_load_environment, env_var_name).extend(env_var_val)
                    else:
                        setattr(self.module_load_environment, env_var_name, env_var_val)

        return super().make_module_step(*args, **kwargs)

    def make_module_extra(self, *args, **kwargs):
        """Set extra stuff in module file, e.g. $EBROOT*, $EBVERSION*, etc."""
        if not self.altroot and not self.altversion:
            # check for altroot and altversion (needed here for a module only build)
            self.altroot, self.altversion = self.get_altroot_and_altversion()
        if 'altroot' not in kwargs:
            kwargs['altroot'] = self.altroot
        if 'altversion' not in kwargs:
            kwargs['altversion'] = self.altversion
        return super().make_module_extra(*args, **kwargs)

    def sanity_check_step(self, *args, **kwargs):
        """
        If component sanity checks are enabled, run sanity checks for the desired components listed.
        If nothing is being installed, just being able to load the (fake) module is sufficient
        """
        if self.cfg['exts_list'] or self.cfg['sanity_check_paths'] or self.cfg['sanity_check_commands']:
            super().sanity_check_step(*args, **kwargs)
        else:
            self.log.info("Testing loading of module '%s' by means of sanity check" % self.full_mod_name)
            fake_mod_data = self.load_fake_module(purge=True)
            self.log.debug("Cleaning up after testing loading of module")
            self.clean_up_fake_module(fake_mod_data)

        # run sanity checks for specific components
        cnt = len(self.comp_cfgs_sanity_check)
        for idx, comp in enumerate(self.comp_cfgs_sanity_check):
            comp_name, comp_ver = comp.cfg['name'], comp.cfg['version']
            print_msg("sanity checking bundle component %s v%s (%i/%i)...", comp_name, comp_ver, idx + 1, cnt)
            self.log.info("Starting sanity check step for component %s v%s", comp_name, comp_ver)

            comp.run_step('sanity_check', [lambda x: x.sanity_check_step])



##
# Copyright 2015-2025 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for creating a module that loads the build
environment flags for the current toolchain

@author: Alan O'Cais (Juelich Supercomputing Centre)
"""
import os

from easybuild.easyblocks.generic.bundle import Bundle
from easybuild.tools.filetools import copy_dir
from easybuild.tools.toolchain.toolchain import RPATH_WRAPPERS_SUBDIR


class BuildEnv(Bundle):
    """
    Build environment of toolchain: only generate module file
    """

    def prepare_step(self, *args, **kwargs):
        """
        Custom prepare step for buildenv: export rpath wrappers if they are being used
        """
        # passed to toolchain.prepare to specify location for RPATH wrapper scripts (if RPATH linking is enabled)
        self.rpath_wrappers_dir = self.builddir

        super(BuildEnv, self).prepare_step(*args, **kwargs)

    def install_step(self, *args, **kwargs):
        """
        Custom install step for buildenv: copy RPATH wrapper scripts to install dir, if desired
        """
        super(BuildEnv, self).install_step(*args, **kwargs)

        # copy RPATH wrapper scripts to install directory (if they exist)
        wrappers_dir = os.path.join(self.rpath_wrappers_dir, RPATH_WRAPPERS_SUBDIR)
        if os.path.exists(wrappers_dir):
            self.rpath_wrappers_dir = os.path.join(self.installdir, 'bin')
            copy_dir(wrappers_dir, os.path.join(self.rpath_wrappers_dir, RPATH_WRAPPERS_SUBDIR))

    def make_module_extra(self):
        """Add all the build environment variables."""
        txt = super().make_module_extra()

        # include environment variables defined for (non-system) toolchain
        if not self.toolchain.is_system_toolchain():
            for key, val in sorted(self.toolchain.vars.items()):
                txt += self.module_generator.set_environment(key, val)

        self.log.debug(f"make_module_extra added this: {txt}")
        return txt

    def make_module_step(self, fake=False):
        """Specify correct bin directories for buildenv installation."""
        wrappers_dir = os.path.join(self.rpath_wrappers_dir, RPATH_WRAPPERS_SUBDIR)
        if os.path.exists(wrappers_dir):
            self.module_load_environment.PATH = [os.path.join(wrappers_dir, d) for d in os.listdir(wrappers_dir)]

        return super(BuildEnv, self).make_module_step(fake=fake)



##
# Copyright 2012-2025 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBlock for installing VSCode and code-cli, implemented as an easyblock
@author: Alan O'Cais (Juelich Supercomputing Centre)
@author: Alex Domingo (Vrije Universiteit Brussel)
"""

from easybuild.easyblocks.generic.tarball import Tarball
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.systemtools import AARCH64, X86_64, get_cpu_architecture


class EB_VSCode(Tarball):
    """
    Support for installing VSCode and code-cli.
    """

    def __init__(self, *args, **kwargs):
        """ Init the easyblock adding a new mapped_arch template var """
        myarch = get_cpu_architecture()
        if myarch == X86_64:
            self.mapped_arch = 'x64'
        elif myarch == AARCH64:
            self.mapped_arch = 'arm64'
        else:
            raise EasyBuildError(f"Architecture {myarch} is not supported for {self.name} on EasyBuild")

        super().__init__(*args, **kwargs)

        self.cfg.template_values['mapped_arch'] = self.mapped_arch
        self.cfg.generate_template_values()

        # installation type: supports VSCode (default) and code-cli
        self.install_type = 'vscode'
        if self.name == 'code-cli':
            self.install_type = 'code-cli'

        # location of VSCode executables:
        bin_path = {
            'vscode': 'bin',  # {installdir}/bin
            'code-cli': '',  # {installdir}
        }
        try:
            self.module_load_environment.PATH = bin_path[self.install_type]
        except KeyError as err:
            raise EasyBuildError(f"Unknown binary location for {self.name} in VSCode easyblock") from err

    def sanity_check_step(self):
        """Custom sanity check for VSCode and code-cli."""
        vscode_paths = {
            'vscode': {
                'files': ['bin/code', 'bin/code-tunnel', 'code'],
                'dirs': ['locales', 'resources'],
            },
            'code-cli': {
                'files': ['code'],
                'dirs': [],
            },
        }
        try:
            custom_paths = vscode_paths[self.install_type]
        except KeyError as err:
            raise EasyBuildError(f"Unknown sanity checks for {self.name} in VSCode easyblock") from err

        custom_commands = ["code --version"]

        return super().sanity_check_step(custom_paths=custom_paths, custom_commands=custom_commands)



#!/usr/bin/env python
##
# Copyright 2009-2025 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for installing Cargo packages (Rust lang package system)

@author: Mikael Oehman (Chalmers University of Technology)
@author: Alex Domingo (Vrije Universiteit Brussel)
@author: Alexander Grund (TU Dresden)
"""

import os
import re
from glob import glob

import easybuild.tools.environment as env
import easybuild.tools.systemtools as systemtools
from easybuild.framework.easyconfig import CUSTOM
from easybuild.framework.extensioneasyblock import ExtensionEasyBlock
from easybuild.tools.build_log import EasyBuildError, print_warning
from easybuild.tools.config import build_option
from easybuild.tools.filetools import CHECKSUM_TYPE_SHA256, compute_checksum, extract_file, mkdir, move_file
from easybuild.tools.filetools import read_file, write_file, which
from easybuild.tools.run import run_shell_cmd
from easybuild.tools.toolchain.compiler import OPTARCH_GENERIC

CRATESIO_SOURCE = "https://crates.io/api/v1/crates"

CONFIG_TOML_SOURCE_VENDOR = """
[source.vendored-sources]
directory = "{vendor_dir}"

[source.crates-io]
replace-with = "vendored-sources"

"""

CONFIG_TOML_SOURCE_GIT = """
[source."{url}?rev={rev}"]
git = "{url}"
rev = "{rev}"
replace-with = "vendored-sources"
"""

CONFIG_TOML_SOURCE_GIT_BRANCH = """
[source."{url}?rev={rev}"]
git = "{url}"
rev = "{rev}"
branch = "{branch}"
replace-with = "vendored-sources"
"""

CONFIG_TOML_SOURCE_GIT_WORKSPACE = """
[source."real-{url}?rev={rev}"]
directory = "{workspace_dir}"

[source."{url}?rev={rev}"]
git = "{url}"
rev = "{rev}"
replace-with = "real-{url}?rev={rev}"

"""

CARGO_CHECKSUM_JSON = '{{"files": {{}}, "package": "{checksum}"}}'


def get_workspace_members(crate_dir):
    """Find all members of a cargo workspace in crate_dir.

    (Minimally) parse the Cargo.toml file.
    If it is a workspace return all members (subfolder names).
    Otherwise return None.
    """
    cargo_toml = os.path.join(crate_dir, 'Cargo.toml')

    # We are looking for this:
    # [workspace]
    # members = [
    # "reqwest-middleware",
    # "reqwest-tracing",
    # "reqwest-retry",
    # ]

    lines = [line.strip() for line in read_file(cargo_toml).splitlines()]
    try:
        start_idx = lines.index('[workspace]')
    except ValueError:
        return None
    # Find "members = [" and concatenate the value, stop at end of section or file
    member_str = None
    for line in lines[start_idx + 1:]:
        if line.startswith('#'):
            continue  # Skip comments
        if re.match(r'\[\w+\]', line):
            break  # New section
        if member_str is None:
            m = re.match(r'members\s+=\s+\[', line)
            if m:
                member_str = line[m.end():]
        else:
            member_str += line
        # Stop if we reach the end of the list
        if member_str is not None and member_str.endswith(']'):
            member_str = member_str[:-1]
            break
    if member_str is None:
        raise EasyBuildError('Failed to find members in %s', cargo_toml)
    # Split at commas after removing possibly trailing ones and remove the quotes
    members = [member.strip().strip('"') for member in member_str.rstrip(',').split(',')]
    # Sanity check that we didn't pick up anything unexpected
    invalid_members = [member for member in members if not re.match(r'(\w|-)+', member)]
    if invalid_members:
        raise EasyBuildError('Failed to parse %s: Found seemingly invalid members: %s',
                             cargo_toml, ', '.join(invalid_members))
    return [os.path.join(crate_dir, m) for m in members]


def get_checksum(src, log):
    """Get the checksum from an extracted source"""
    checksum = src['checksum']
    if isinstance(checksum, dict):
        try:
            checksum = checksum[src['name']]
        except KeyError:
            log.warning('No checksum for %s in %s', checksum, src['name'])
            checksum = None
    return checksum


class Cargo(ExtensionEasyBlock):
    """Support for installing Cargo packages (Rust)"""

    @staticmethod
    def extra_options(extra_vars=None):
        """Define extra easyconfig parameters specific to Cargo"""
        extra_vars = ExtensionEasyBlock.extra_options(extra_vars)
        extra_vars.update({
            'enable_tests': [True, "Enable building of tests", CUSTOM],
            'offline': [True, "Build offline", CUSTOM],
            'lto': [None, "Override default LTO flag ('fat', 'thin', 'off')", CUSTOM],
            'crates': [[], "List of (crate, version, [repo, rev]) tuples to use", CUSTOM],
        })

        return extra_vars

    @staticmethod
    def crate_src_filename(pkg_name, pkg_version, _url=None, rev=None):
        """Crate tarball filename based on package name, version and optionally git revision"""
        filename = [pkg_name, pkg_version]
        filename_ext = '.tar.gz'

        if rev is not None:
            # sources from a git repo
            filename.append(rev[:8])  # append short commit hash
            filename_ext = '.tar.xz'  # use a reproducible archive format

        return '-'.join(filename) + filename_ext

    @staticmethod
    def crate_download_filename(pkg_name, pkg_version):
        """Crate download filename based on package name and version"""
        return f"{pkg_name}/{pkg_version}/download"

    def rustc_optarch(self):
        """Determines what architecture to target.
        Translates GENERIC optarch, and respects rustc specific optarch.
        General optarchs are ignored as there is no direct translation.
        """
        generic = '-C target-cpu=generic'
        if systemtools.get_cpu_architecture() == systemtools.X86_64:
            generic = '-C target-cpu=x86-64'

        optimal = '-C target-cpu=native'

        optarch = build_option('optarch')
        if optarch:
            if isinstance(optarch, dict):
                if 'rustc' in optarch:
                    rust_optarch = optarch['rustc']
                    if rust_optarch == OPTARCH_GENERIC:
                        return generic
                    else:
                        return '-' + rust_optarch
                self.log.info(f"Given 'optarch' has no specific information on rustc, so using {optimal}")
            elif optarch == OPTARCH_GENERIC:
                return generic
            else:
                self.log.warning(f"Ignoring 'optarch' because there is no translation for rustc, so using {optimal}")

        return optimal

    def __init__(self, *args, **kwargs):
        """Constructor for Cargo easyblock."""
        super().__init__(*args, **kwargs)
        self.cargo_home = os.path.join(self.builddir, '.cargo')
        self.set_cargo_vars()

        # Populate sources from "crates" list of tuples
        sources = []
        for crate_info in self.crates:
            if len(crate_info) == 2:
                sources.append({
                    'download_filename': self.crate_download_filename(*crate_info),
                    'filename': self.crate_src_filename(*crate_info),
                    'source_urls': [CRATESIO_SOURCE],
                    'alt_location': 'crates.io',
                })
            else:
                crate, version, repo, rev = crate_info
                url, repo_name = repo.rsplit('/', maxsplit=1)
                if repo_name.endswith('.git'):
                    repo_name = repo_name[:-4]
                sources.append({
                    'git_config': {'url': url, 'repo_name': repo_name, 'commit': rev},
                    'filename': self.crate_src_filename(crate, version, rev=rev),
                })

        # copy EasyConfig instance before we make changes to it
        self.cfg = self.cfg.copy()

        self.cfg.update('sources', sources)

    def set_cargo_vars(self):
        """Set environment variables for Rust compilation and Cargo"""
        rustc_optarch = self.rustc_optarch()
        gcc = which('gcc')  # makes sure gcc wrapper is used in case of rpath linking.

        env.setvar('CARGO_HOME', self.cargo_home)
        env.setvar('RUSTC', 'rustc')
        env.setvar('RUSTDOC', 'rustdoc')
        env.setvar('RUSTFMT', 'rustfmt')
        env.setvar('RUSTFLAGS', f'{rustc_optarch} -C linker={gcc}')
        env.setvar('RUST_LOG', 'DEBUG')
        env.setvar('RUST_BACKTRACE', '1')

    @property
    def crates(self):
        """Return the crates as defined in the EasyConfig"""
        return self.cfg['crates']

    def load_module(self, *args, **kwargs):
        """(Re)set environment variables after loading module file.

        Required here to ensure the variables are defined for stand-alone installations and extensions,
        because the environment is reset to the initial environment right before loading the module.
        """
        super().load_module(*args, **kwargs)
        self.set_cargo_vars()

    def extract_step(self):
        """
        Unpack the source files and populate them with required .cargo-checksum.json if offline
        """
        self.vendor_dir = os.path.join(self.builddir, 'easybuild_vendor')
        mkdir(self.vendor_dir)
        # Sources from git repositories might contain multiple crates/folders in a so-called "workspace".
        # If we put such a workspace into the vendor folder, cargo fails with
        # "found a virtual manifest at [...]Cargo.toml instead of a package manifest".
        # Hence we put those in a separate folder and only move "regular" crates into the vendor folder.
        self.git_vendor_dir = os.path.join(self.builddir, 'easybuild_vendor_git')
        mkdir(self.git_vendor_dir)

        vendor_crates = {self.crate_src_filename(*crate): crate for crate in self.crates}
        # Track git sources for building the cargo config and avoiding duplicated folders
        git_sources = {}

        for src in self.src:
            # Check if the source is a vendored crate
            is_vendor_crate = src['name'] in vendor_crates
            if is_vendor_crate:
                # Store crate for later
                src['crate'] = vendor_crates[src['name']]
                crate_name = src['crate'][0]

            # Check for git crates, `git_key` will be set to a true-ish value for those
            if not is_vendor_crate or len(src['crate']) == 2:
                git_key = None
            else:
                git_key = src['crate'][2:]
                git_repo, rev = git_key
                self.log.debug("Sources of %s(%s) belong to git repo: %s rev %s",
                               crate_name, src['name'], git_repo, rev)
                # Do a sanity check that sources for the same repo and revision are the same
                try:
                    previous_source = git_sources[git_key]
                except KeyError:
                    git_sources[git_key] = src
                else:
                    previous_checksum = get_checksum(previous_source, self.log)
                    current_checksum = get_checksum(src, self.log)
                    if previous_checksum and current_checksum and previous_checksum != current_checksum:
                        raise EasyBuildError("Sources for the same git repository need to be identical. "
                                             "Mismatch found for %s rev %s in %s (checksum: %s) vs %s (checksum: %s)",
                                             git_repo, rev, previous_source['name'], previous_checksum,
                                             src['name'], current_checksum)
                    self.log.info("Source %s already extracted to %s by %s. Skipping extraction.",
                                  src['name'], previous_source['finalpath'], previous_source['name'])
                    src['finalpath'] = previous_source['finalpath']
                    continue

            # Extract dependency crates into vendor subdirectory, separate from sources of main package
            extraction_dir = self.builddir
            if is_vendor_crate:
                extraction_dir = self.git_vendor_dir if git_key else self.vendor_dir

            self.log.info("Unpacking source of %s", src['name'])
            existing_dirs = set(os.listdir(extraction_dir))
            src_dir = extract_file(src['path'], extraction_dir, cmd=src['cmd'],
                                   extra_options=self.cfg['unpack_options'], change_into_dir=False)
            new_extracted_dirs = set(os.listdir(extraction_dir)) - existing_dirs

            if len(new_extracted_dirs) == 0:
                # Extraction went wrong
                raise EasyBuildError("Unpacking sources of '%s' failed", src['name'])
            # Expected crate tarball with 1 folder
            # TODO: properly handle case with multiple extracted folders
            # this is currently in a grey area, might still be used by cargo
            if len(new_extracted_dirs) == 1:
                src_dir = os.path.join(extraction_dir, new_extracted_dirs.pop())
                self.log.debug("Unpacked sources of %s into: %s", src['name'], src_dir)

                if is_vendor_crate and self.cfg['offline']:
                    # Create checksum file for extracted sources required by vendored crates

                    # By default there is only a single crate
                    crate_dirs = [src_dir]
                    # For git sources determine the folders that contain crates by taking workspaces into account
                    if git_key:
                        member_dirs = get_workspace_members(src_dir)
                        if member_dirs:
                            crate_dirs = member_dirs

                    try:
                        checksum = src[CHECKSUM_TYPE_SHA256]
                    except KeyError:
                        checksum = compute_checksum(src['path'], checksum_type=CHECKSUM_TYPE_SHA256)
                    for crate_dir in crate_dirs:
                        self.log.info('creating .cargo-checksums.json file for %s', os.path.basename(crate_dir))
                        chkfile = os.path.join(src_dir, crate_dir, '.cargo-checksum.json')
                        write_file(chkfile, CARGO_CHECKSUM_JSON.format(checksum=checksum))
                    # Move non-workspace git crates to the vendor folder
                    if git_key and member_dirs is None:
                        src_dir = os.path.join(self.vendor_dir, os.path.basename(crate_dirs[0]))
                        self.log.debug('Moving crate %s without workspaces to vendor folder', crate_name)
                        move_file(crate_dirs[0], src_dir)

            src['finalpath'] = src_dir

        self._setup_offline_config(git_sources)

    def _setup_offline_config(self, git_sources):
        """
        Setup the configuration required for offline builds
        :param git_sources: dict mapping (git_repo, rev) to extracted source
        """
        if not self.cfg['offline']:
            return
        self.log.info("Setting up vendored crates for offline operation")
        self.log.debug("Writting config.toml entry for vendored crates from crate.io")
        config_toml = os.path.join(self.cargo_home, 'config.toml')
        # Replace crates-io with vendored sources using build dir wide toml file in CARGO_HOME
        write_file(config_toml, CONFIG_TOML_SOURCE_VENDOR.format(vendor_dir=self.vendor_dir))

        # Tell cargo about the vendored git sources to avoid it failing with:
        # Unable to update https://github.com/[...]
        # can't checkout from 'https://github.com/[...]]': you are in the offline mode (--offline)

        for (git_repo, rev), src in git_sources.items():
            crate_name = src['crate'][0]
            src_dir = src['finalpath']
            if os.path.dirname(src_dir) == self.vendor_dir:
                # Non-workspace sources are in vendor_dir
                git_branch = self._get_crate_git_repo_branch(crate_name)
                template = CONFIG_TOML_SOURCE_GIT_BRANCH if git_branch else CONFIG_TOML_SOURCE_GIT
                self.log.debug(f"Writing config.toml entry for git repo: {git_repo} branch {git_branch}, rev {rev}")
                write_file(
                    config_toml,
                    template.format(url=git_repo, rev=rev, branch=git_branch),
                    append=True
                )
            else:
                self.log.debug("Writing config.toml entry for git repo: %s rev %s", git_repo, rev)
                # Workspace sources stay in their own separate folder.
                # We cannot have a `directory = "<dir>"` entry where a folder containing a workspace is inside
                write_file(
                    config_toml,
                    CONFIG_TOML_SOURCE_GIT_WORKSPACE.format(url=git_repo, rev=rev, workspace_dir=src_dir),
                    append=True
                )

        # Use environment variable since it would also be passed along to builds triggered via python packages
        env.setvar('CARGO_NET_OFFLINE', 'true')

    def _get_crate_git_repo_branch(self, crate_name):
        """
        Find the dependency definition for given crate in all Cargo.toml files of sources
        Return branch target for given crate_name if any
        """
        # Search all Cargo.toml files in main source and vendored crates
        cargo_toml_files = []
        for cargo_source_dir in (self.src[0]['finalpath'], self.vendor_dir, self.git_vendor_dir):
            cargo_toml_files.extend(glob(os.path.join(cargo_source_dir, '**', 'Cargo.toml'), recursive=True))

        if not cargo_toml_files:
            raise EasyBuildError("Cargo.toml file not found in sources")

        self.log.debug(
            f"Searching definition of crate '{crate_name}' in the following files: {', '.join(cargo_toml_files)}"
        )

        git_repo_spec = re.compile(re.escape(crate_name) + r"\s*=\s*{([^}]*)}", re.M)
        git_branch_spec = re.compile(r'branch\s*=\s*"([^"]*)"', re.M)

        for cargo_toml in cargo_toml_files:
            git_repo_crate = git_repo_spec.search(read_file(cargo_toml))
            if git_repo_crate:
                self.log.debug(f"Found specification in {cargo_toml} for crate '{crate_name}': " +
                               git_repo_crate.group())
                git_repo_crate_contents = git_repo_crate.group(1)
                git_branch_crate = git_branch_spec.search(git_repo_crate_contents)
                if git_branch_crate:
                    self.log.debug(f"Found git branch requirement for crate '{crate_name}': " +
                                   git_branch_crate.group())
                    return git_branch_crate.group(1)

        return None

    def configure_step(self):
        """Empty configuration step."""
        pass

    @property
    def profile(self):
        return 'debug' if self.toolchain.options.get('debug', None) else 'release'

    def build_step(self):
        """Build with cargo"""
        parallel = ''
        if self.cfg.parallel > 1:
            parallel = f"-j {self.cfg.parallel}"

        tests = ''
        if self.cfg['enable_tests']:
            tests = "--tests"

        lto = ''
        if self.cfg['lto'] is not None:
            lto = f'--config profile.{self.profile}.lto="{self.cfg["lto"]}"'

        run_shell_cmd('rustc --print cfg')  # for tracking in log file
        cmd = ' '.join([
            self.cfg['prebuildopts'],
            'cargo build',
            '--profile=' + self.profile,
            lto,
            tests,
            parallel,
            self.cfg['buildopts'],
        ])
        run_shell_cmd(cmd)

    def test_step(self):
        """Test with cargo"""
        if self.cfg['enable_tests']:
            cmd = ' '.join([
                self.cfg['pretestopts'],
                'cargo test',
                '--profile=' + self.profile,
                self.cfg['testopts'],
            ])
            run_shell_cmd(cmd)

    def install_step(self):
        """Install with cargo"""
        cmd = ' '.join([
            self.cfg['preinstallopts'],
            'cargo install',
            '--profile=' + self.profile,
            '--root=' + self.installdir,
            '--path=.',
            self.cfg['installopts'],
        ])
        run_shell_cmd(cmd)


def generate_crate_list(sourcedir):
    """Helper for generating crate list"""
    from urllib.parse import parse_qs, urlsplit  # pylint: disable=import-outside-toplevel

    import toml  # pylint: disable=import-outside-toplevel

    cargo_toml = toml.load(os.path.join(sourcedir, 'Cargo.toml'))

    try:
        cargo_lock = toml.load(os.path.join(sourcedir, 'Cargo.lock'))
    except FileNotFoundError as err:
        print("\nNo Cargo.lock file found. Generate one with 'cargo generate-lockfile'\n")
        raise err

    try:
        app_name = cargo_toml['package']['name']
    except KeyError:
        app_name = os.path.basename(os.path.abspath(sourcedir))
        print_warning('Did not find a [package] name= entry. Assuming it is the folder name: ' + app_name)

    app_in_cratesio = False
    crates = []
    other_crates = []
    for dep in cargo_lock['package']:
        name = dep['name']
        version = dep['version']
        try:
            source_url = dep['source']
        except KeyError:
            other_crates.append((name, version))
            continue
        if name == app_name:
            app_in_cratesio = True  # exclude app itself, needs to be first in crates list or taken from pypi
        else:
            if source_url == 'registry+https://github.com/rust-lang/crates.io-index':
                crates.append((name, version))
            else:
                # Lock file has revision and branch in the url
                url = re.sub(r'^(registry|git)\+', '', source_url)  # Strip prefix if present
                parsed_url = urlsplit(url)
                url = re.split('[#?]', url, maxsplit=1)[0]  # Remove query and fragment
                rev = parsed_url.fragment
                if not rev:
                    raise ValueError(f"Revision not found in URL {url}")
                qs = parse_qs(parsed_url.query)
                rev_qs = qs.get('rev', [None])[0]
                if rev_qs is not None and rev_qs != rev:
                    raise ValueError(f"Found different revision in query of URL {url}: {rev_qs} (expected: {rev})")
                crates.append((name, version, url, rev))
    return app_in_cratesio, crates, other_crates


def main():
    import sys  # pylint: disable=import-outside-toplevel
    if len(sys.argv) != 2:
        print('Expected path to folder containing Cargo.[toml,lock]')
        sys.exit(1)
    app_in_cratesio, crates, other = generate_crate_list(sys.argv[1])
    print('Other crates (no source in Cargo.lock):', other)
    if app_in_cratesio or crates:
        print('crates = [')
        if app_in_cratesio:
            print('    (name, version),')
        for crate_info in sorted(crates):
            print("    %s," % str(crate_info))
        print(']')


if __name__ == '__main__':
    main()



##
# Copyright 2009-2025 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for installing datasets

@author: Samuel Moors (Vrije Universiteit Brussel)
"""
import os

from easybuild.framework.easyblock import EasyBlock
from easybuild.easyblocks.generic.binary import Binary
from easybuild.framework.easyconfig.default import CUSTOM
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.filetools import compute_checksum, create_index, is_readable, mkdir, move_file, remove_file
from easybuild.tools.filetools import symlink
from easybuild.tools.utilities import trace_msg


class Dataset(Binary):
    """Support for installing datasets"""

    @staticmethod
    def extra_options(extra_vars=None):
        """Extra easyconfig parameters specific to Data easyblock."""
        extra_vars = EasyBlock.extra_options(extra_vars)
        extra_vars.update({
            'extract_sources': [True, "Whether or not to extract data sources", CUSTOM],
            'data_install_path': [None, "Custom installation path for datasets", CUSTOM],
            'cleanup_data_sources': [False, "Whether or not to delete the data sources after installation", CUSTOM]
        })
        return extra_vars

    def __init__(self, *args, **kwargs):
        """Initialize Dataset-specific variables."""
        super().__init__(*args, **kwargs)

        if self.cfg['sources']:
            raise EasyBuildError(
                "Easyconfig parameter 'sources' is not supported for this EasyBlock. Use 'data_sources' instead.")

        if self.cfg['data_install_path']:
            self.installdir = self.cfg['data_install_path']

        # extract/copy sources directly into installation directory
        self.build_in_installdir = True

    def install_step(self):
        """No install step, datasets are extracted directly into installdir"""
        pass

    def post_processing_step(self):
        """Add files to object_storage, remove duplicates, add symlinks"""
        trace_msg('adding files to object_storage...')

        # creating object storage at root of software name to reuse identical files in different versions
        object_storage = os.path.join(os.pardir, 'object_storage')
        datafiles = create_index(os.curdir)

        for datafile in datafiles:
            cks = compute_checksum(datafile, checksum_type='sha256')
            # using puppet-style object store, for example this checksum:
            # 00b68cbca8fe75a121e857359191f481d2e1262ce7c9998e9980fdb35c144733
            # is stored at:
            # 0/0/b/6/8/c/b/c/00b68cbca8fe75a121e857359191f481d2e1262ce7c9998e9980fdb35c144733
            objstor_file = os.path.join(object_storage, os.sep.join(list(cks[:8])), cks)
            mkdir(os.path.dirname(objstor_file), parents=True)
            if is_readable(objstor_file):
                remove_file(datafile)
            else:
                move_file(datafile, objstor_file)
            # use relative paths for symlinks to easily relocate data installations later on if needed
            symlink(objstor_file, datafile, use_abspath_source=False)
            self.log.debug(f"Created symlink {datafile} to {objstor_file}")

    def cleanup_step(self):
        """Cleanup sources after installation"""
        if self.cfg['cleanup_data_sources']:
            for src in self.src:
                self.log.info(f"Removing data source {src['name']}")
                remove_file(src['path'])
        super().cleanup_step()



##
# Copyright 2009-2025 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for building and installing freetype, implemented as an easyblock

@author: Kenneth Hoste (Ghent University)
"""

from easybuild.easyblocks.generic.configuremake import ConfigureMake
from easybuild.tools.modules import MODULE_LOAD_ENV_HEADERS
from easybuild.tools.systemtools import get_shared_lib_ext


class EB_freetype(ConfigureMake):
    """Support for building/installing freetype."""

    def __init__(self, *args, **kwargs):
        """Initialisation of custom class variables for freetype."""
        super().__init__(*args, **kwargs)

        self.maj_ver = self.version.split('.')[0]

        self.module_load_environment.set_alias_vars(MODULE_LOAD_ENV_HEADERS, [f'include/freetype{self.maj_ver}'])

    def sanity_check_step(self):
        """Custom sanity check for freetype."""
        custom_paths = {
            'files': ['bin/freetype-config', 'lib/libfreetype.a', 'lib/libfreetype.%s' % get_shared_lib_ext(),
                      'lib/pkgconfig/freetype%s.pc' % self.maj_ver],
            'dirs': ['include/freetype%s' % self.maj_ver],
        }
        super().sanity_check_step(custom_paths=custom_paths)
